#include "stdafx.h"
#include <stddef.h>
#include "SyncRun.h"
#include "misc.h"

SyncRun::SyncRun(void)
	: sharedSize(0)
	, sharedHandle(NULL)
	, sharedPointer(NULL)
	, runMode(0)
	, slaveWnd(NULL)
	, waitForSlaveClose(FALSE)
{
	memset( sharedIdent, 0, lengthof(sharedIdent) );
	memset( slavePath, 0, lengthof(slavePath) );
}

SyncRun::~SyncRun(void)
{
	if ( sharedPointer ) {
		UnmapViewOfFile( sharedPointer );
		sharedPointer = NULL;
	}
	if ( sharedHandle ) {
		CloseHandle( sharedHandle );
		sharedHandle = NULL;
	}
}

BOOL SyncRun::Init(HWND hWnd, TCHAR *ident, DWORD size, TCHAR *path)
{
	BOOL bRet = FALSE;

	sharedSize = size;
	if ( path ) {
		_tcscpy_s( slavePath, _MAX_PATH, path );
	}
	if ( ident ) {
		_tcscpy_s( sharedIdent, lengthof(sharedIdent), ident );
	}

	do {
		sharedHandle = OpenFileMapping(
				FILE_MAP_READ,
				FALSE,
				sharedIdent );

		if ( sharedHandle == NULL ) {

			sharedHandle = CreateFileMapping(
					INVALID_HANDLE_VALUE,
					NULL,
					PAGE_READWRITE,
					0,
					sharedSize + offsetof( struct _SHARED_STRUCT, customStruct ),
					sharedIdent );

			if ( sharedHandle == NULL ) {
				break;
			}

			sharedPointer = (SHARED_STRUCT*)MapViewOfFile( sharedHandle, FILE_MAP_ALL_ACCESS, 0, 0, 0 );
			if ( sharedPointer == NULL ) {
				break;
			}

			sharedPointer->masterWnd = (DWORD)hWnd;
			sharedPointer->ref = 0;

			if ( path != NULL ) {
				SHELLEXECUTEINFO sei = {0};
				sei.cbSize = sizeof(SHELLEXECUTEINFO);
				sei.nShow = SW_NORMAL;
				sei.lpFile = slavePath;
				sei.fMask = SEE_MASK_FLAG_NO_UI;

				if ( ShellExecuteEx( &sei ) ) {
					runMode = RUN_MODE_MASTER;
				}
				else {
					runMode = RUN_MODE_SINGLE;
				}
			}
			else {
				runMode = RUN_MODE_SINGLE;
			}

			bRet = TRUE;
		}
		else {

			sharedPointer = (SHARED_STRUCT*)MapViewOfFile( sharedHandle, FILE_MAP_READ, 0, 0, 0 );
			if ( sharedPointer == NULL ) {
				break;
			}

			if ( sharedPointer->ref != 0 ) {
				break;
			}

			if ( !PostMessage( (HWND)sharedPointer->masterWnd, WM_NULL, 0, 0 ) ) {
				break;
			}

			runMode = RUN_MODE_SLAVE;

			PostMessage( (HWND)sharedPointer->masterWnd, WM_SHARED_HWND, (WPARAM)hWnd, 1 );

			bRet = TRUE;
		}

	} while( FALSE );

	if ( !bRet ) {
		if ( sharedPointer ) {
			UnmapViewOfFile( sharedPointer );
			sharedPointer = NULL;
		}
		if ( sharedHandle ) {
			CloseHandle( sharedHandle );
			sharedHandle = NULL;
		}
	}

	return bRet;
}

INT_PTR CALLBACK SyncRun::TranslateMessage( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch( uMsg )
	{
	case WM_SHARED_HWND:
		if ( runMode != RUN_MODE_SLAVE ) {
			switch( lParam )
			{
			case 1:
				slaveWnd = (HWND)wParam;
				sharedPointer->ref++;
				if ( runMode != RUN_MODE_MASTER ) {
					runMode = RUN_MODE_MASTER;
					PostMessage( hWnd, WM_SHARED_MODE, 0, 0 );
				}
				break;

			case 2:
				if ( sharedPointer->ref > 0 ) {
					sharedPointer->ref--;
				}
				if ( ( sharedPointer->ref == 0 ) && ( runMode != RUN_MODE_SINGLE ) ) {
					runMode = RUN_MODE_SINGLE;
					PostMessage( hWnd, WM_SHARED_MODE, 0, 0 );
				}
				if ( waitForSlaveClose ) {
					DestroyWindow( hWnd );
				}
				break;
			}
			return TRUE;
		}
		break;

	case WM_SHARED_CLOSE:
		PostMessage( (HWND)sharedPointer->masterWnd, WM_SHARED_HWND, (WPARAM)hWnd, 2 );
		DestroyWindow( hWnd );
		return TRUE;
	}
	return 0;	
}

BOOL SyncRun::SharedUpdate(void)
{
	if ( ( runMode == RUN_MODE_MASTER ) && ( slaveWnd ) ) {
		return PostMessage( slaveWnd, WM_SHARED_UPDATE, 0, 0 );
	}
	return FALSE;
}

void SyncRun::CloseApp( HWND hWnd )
{
	switch( runMode )
	{
	case RUN_MODE_SINGLE:
		DestroyWindow( hWnd );
		break;

	case RUN_MODE_MASTER:
		if ( ( slaveWnd ) && ( sharedPointer->ref > 0 ) ) {
			waitForSlaveClose = TRUE;
			if ( !PostMessage( slaveWnd, WM_SHARED_CLOSE, 0, 0 ) ) {
				DestroyWindow( hWnd );
			}
		}
		else {
			DestroyWindow( hWnd );
		}
		break;

	case RUN_MODE_SLAVE:
		PostMessage( (HWND)sharedPointer->masterWnd, WM_SHARED_HWND, (WPARAM)hWnd, 2 );
		DestroyWindow( hWnd );
		break;
	}
}

void *SyncRun::GetSharedPointer(void)
{
	return sharedPointer ? (LPVOID)( (LPBYTE)sharedPointer + offsetof( struct _SHARED_STRUCT, customStruct ) ) : NULL;
}
